#ifndef EXTCLS_HPP_
#define EXTCLS_HPP_
#define extcls_hpp_version "$Revision: 1.24 $"

#include "ThreadGlobals.hpp"
#include "Types.hpp"
#include "Settings.hpp"
#include <vector>
#include "Hash.hpp"
#include "VecOfLists.hpp"
#include "AssignInfos.hpp"
#include "Queue.hpp"


typedef uint32_t signt_t;



class ExtCls{
	Literal		one_lit;

	uint32_t	wst 	: 1; // clause is sat / waste - may still be linked by occurrence lists
	uint32_t	lrnt	: 1; // clause is redundant / learnt
	uint32_t	queued	: 1; // enqueued in subsum_q
	uint32_t 	flg		: 1; //temporarily added as cross product in var elimination
	uint32_t	asym	: 1; // true if asym check has to be done
	uint32_t	siz		: 27; // size of linked clause

	ClsP 		clp;
	signt_t 	sig; //signature for subsumption checks

	inline ExtCls(uint32_t size, const bool learnt, const bool flag, const ClsP& cp, signt_t signat,
				const Literal& lit, const bool qd, const bool do_assym):
			one_lit(lit),wst(0),lrnt(learnt),queued(qd),flg(flag),asym(do_assym),siz(size),clp(cp),sig(signat){
			asrt(size>1);asrt(!flag || learnt);}
public:
	//gc
	bool		isWaste() const		{return wst;}
	void		doWaste() 			{asrt(wst==0); wst=1;}
	bool		doWasteWasAlve()	{bool ret=wst; wst=1; return !ret;}
	void		undoWaste() 		{asrt(wst==1); wst=0;}

	bool		isMrkd()const		{return flg;}
	void		setMrk(bool b)		{flg=b;}
	void		setAssym(bool b)	{asym=b;}

	uint32_t	size() const		{return siz;}
	void		setSize(uint32_t i)	{asrt(i>2); siz=i;}
	void		setSigSize(const uint32_t& sz,const signt_t& sg)	{asrt(sz>2); asrt(sg!=0);siz=sz; sig=sg;}
	signt_t	getSig() const{return sig;}

	bool		isLearnt() const	{return lrnt;}
	void		setFix()			{lrnt = 0;flg=0;}
	bool		shallAssymChd()const{return asym;}

	//subsumption queue - clause will be checked if it subsumes or bw subsumes any other clause
	//void		setQueued()			{queued=1;}
	//bool		isQueued()const		{return queued;}

	void		setLit(const Literal& lit){one_lit=lit;}
	Literal		getLit() const		{return one_lit;}

	ClsP		getClp()const{return clp;}

	//normal clauses

	static signt_t getHash(const Literal& l){
				return hashBit32(l.toVar().idx());}

	bool			cannotBeSubsumedBy(const uint32_t& sub_sz, const signt_t& sub_sig){
				return (siz < sub_sz) || (~sig & sub_sig) ;}
	bool			mayContainOneOf(const signt_t& lits_msk){
				return sig & lits_msk;}

	//sig_t	getSig()const			{return sig;}
	//void		setSig(sig_t h)		{sig=h;}
	//void		addHashLit(const Literal& l){sig |= getHash(l);}


	friend class ExtClses;
};


struct ECP{ //index into eclses of class ClsInfos
	uint32_t idx;
	ECP(){};
	explicit ECP(uint32_t i):idx(i){}
};
inline bool operator==(const ECP& a, const ECP& b){return a.idx==b.idx;}


class ExtClses{
	VecOfLists<Literal,Literal> 		bins;
	vector<ExtCls>				eclses;

	Queue<ECP,32>				subsum_q;//clses that may subsume other clauses
	Queue<Literal,32>			subsum_bq;//binaries that may subsume other clauses
	BitVector<Literal>			bin_eq;

	bool bins_udt;

	//inline bool		addOneImpl(const Literal& a, const Literal& b);
	void			enqueBin(const Literal& a,const Literal& b){
				if(!bin_eq.isSet(a) && !bin_eq.isSet(b)){bin_eq.setFlag(a); subsum_bq.push_back(a); }}

public:
	ExtClses()		:bins('B'),bins_udt(false){}
	void			ensureLits(const uint32_t l){bins.enlarge(l);bin_eq.ensureSize(l,false);}
	Iter<Literal>	getBinsOf(const Literal& a)const {return bins.getIter(a);}
	WIter<Literal>	getWBinsOf(const Literal& a){return bins.getWIter(a);}
	void			finishBins(const Literal& a, WIter<Literal> wi){bins.finish(a,wi);}
	Range<Literal>  getBinsRangeOf(const Literal& a){return bins.getRange(a);}
	bool			getNextBin(Range<Literal>& r, Literal& out){return bins.getNext(r,out);}
	void			replaceNextBin(Range<Literal>& r, const Literal& n){bins.writeNext(r,n);}
	void			finishBinReplace(const Literal& a, Range<Literal>& r){bins.finishWrite(a,r);}
	void			clearBinsPartOf(const Literal& a){bins.clear(a); LOG2("Cleared all bins of " << a.toDIMACS());}
	bool			noBinsPart(const Literal& a)const{return bins.empty(a);}
	uint32_t		nbrBins(const Literal& a)const{return bins.sizeOf(a);}

	ExtCls&			getECls(const ECP& p){
					asrt(p.idx < eclses.size());
					return eclses[p.idx];}

	void				setLitFor(const ECP& p, const Literal& l){asrt(p.idx < eclses.size());
					eclses[p.idx].setLit(l);}

	void				enqSubTest(const ECP& ecp){
					ExtCls& ec = getECls(ecp);
					if(ec.queued == 0){ec.queued = 1; subsum_q.push_back(ecp);}}

	const ExtCls&	getECls(const ECP& p)const{
					asrt(p.idx < eclses.size());
					return eclses[p.idx];}

	bool 			binsUpdated(){bool ret=bins_udt;bins_udt=false;return ret;}

	uint32_t		maxClses() const{return eclses.size();}

	uint32_t		binsQSize() const{return subsum_bq.size();}
	uint32_t		subQSize() const{return subsum_q.size();}

	inline bool 		dequeueSubLit(Literal& lit);
	inline bool 		dequeueSubCls(ECP& ecp);

	inline void		addBinPart(const Literal& a, const Literal& b){ LOG2("Add bin part " << a.toDIMACS() << " to " << b.toDIMACS());
					bins.add(a,b); enqueBin(a,b); asrt(a != b); asrt(a != b.flipped());}
	inline void		addBin(const Literal& a, const Literal& b);
	inline void		removeBinPart(const Literal& of, const Literal& who);
	inline ECP		addCls(const ClsP& clp, const Literal& lit, const uint32_t size, const bool lrnt,
					const signt_t& sig, const bool enq_sub, const bool tmp, const bool do_assym);

	inline bool		asrtBinBothDirs(const VarVal& vals)const ;
	inline bool		asrtBinContained(const Literal& frst, const Literal& sec)const ;
	inline bool 	asrtNoMrkdClses()const;

};

struct EWtch{
	ClsP 	clp;
	ECP 	extp;
	EWtch(const ClsP& c, const ECP& e):clp(c),extp(e){}
};
inline bool operator==(const EWtch& a, const EWtch& b){return a.clp==b.clp && a.extp==b.extp;}


struct Subsumd{
	ECP who;
	ECP by_whom;
	Subsumd(const ECP& subsumed, const ECP& subsuming):who(subsumed),by_whom(subsuming){}
};



void ExtClses::removeBinPart(const Literal& of, const Literal& who){
	LOG2("Binary in " << of.toDIMACS() << " let die " << who.toDIMACS());
	WIter<Literal> wi = bins.getWIter(of); Literal cand; asrt(of != who);
	while(wi.getNext(cand)){if(cand!=who){wi.keep(cand);}}
	bins.finish(of,wi);
}

void ExtClses::addBin(const Literal& a, const Literal& b){
	LOG2("Add bin " << a.toDIMACS() << " , " << b.toDIMACS());
	bins.add(a,b); asrt(a != b); asrt(a != b.flipped());
	bins.add(b,a);
	bins_udt = true;
	enqueBin(a,b);
}

bool ExtClses::dequeueSubLit(Literal& lit){
	if(subsum_bq.pop_front(lit)){	bin_eq.clear(lit);return true;}
	return false;
}

bool ExtClses::dequeueSubCls(ECP& ecp){
	if(subsum_q.pop_front(ecp)){	getECls(ecp).queued =0;return true;}
	return false;
}

ECP	ExtClses::addCls(const ClsP& clp, const Literal& lit, const uint32_t size, const bool lrnt,
						const signt_t& sig, const bool enq_sub,const bool flg, const bool do_assym){asrt(sig!=0);
	ECP ret(eclses.size());
	eclses	.push_back(ExtCls(size,lrnt,flg,clp,sig,lit,enq_sub,do_assym));
	if(enq_sub){subsum_q.push_back(ret);}
	return 	ret;
}


bool	ExtClses::asrtBinBothDirs(const VarVal& vals)const {
	bool ok = true; Literal cand;
	for(uint32_t i=bins.listCount();i--;){
		Literal lit(i);
		if(vals.litKnwn(lit)) continue;

		Iter<Literal> it = bins.getIter(lit);
		while(it.getNext(cand)){
			if(vals.litUkwn(cand) && !asrtBinContained(cand,lit)){ ok=false;
				LOG0("Contains bin only in one dir: " << lit.toDIMACS() << " to " << cand.toDIMACS());
			}
		}
	}
	LOG1("ExtClses::asrtBinBothDirs is " << (ok?"ok":"bad"));

	return ok;
}

bool	ExtClses::asrtBinContained(const Literal& frst, const Literal& sec)const {
	Iter<Literal> it = bins.getIter(frst); Literal cand;
	while(it.getNext(cand)){
		if(cand==sec) return true;
	}
	return false;
}

bool 	ExtClses::asrtNoMrkdClses()const{
	bool ok = true;
	for(uint32_t i=eclses.size();i--;){
		const ExtCls& ec = eclses[i];
		if(!ec.isWaste() && ec.isMrkd()){ ok = false;
			LOG0("Clause " << i << " is still marked");}
	}
	return ok;
}


/*struct IntPair{
	uint32_t frst;
	uint32_t sec;
	//IntPair(const uint32_t& a, const uint32_t& b):
};*/


/*
bool ExtClses::addOneImpl(const Literal& a, const Literal& b){
	asrt(lit_flags.allCleared()); asrt(!vals.varKnwn(a.toVar()));
	Iter<Literal> at = impls.getIter(a); Literal lit; uint32_t mrk=0;
	while(at.getNext(lit)){asrt(!lit_flags.isSet(lit));lit_flags.setFlag(lit);++mrk;}

	lit_flags.setFlag(a);
	if(!lit_flags.isSet(b)){lit_flags.setFlag(b); impls.add(a,b);}

	Iter<Literal> bt= impls.getIter(b);
	bool unt		= false;
	bool eq			= false;
	while(bt.getNext(lit)){
		unt |= lit_flags.isSet(lit.flipped());
		eq	|= (lit==a);
		if(!lit_flags.isSet(lit)){
			impls.add(a,lit); lit_flags.setFlag(lit);}
	}
	lit_flags.carelessClear(a);
	lit_flags.carelessClear(b);
	while(mrk--&&bt.getNext(lit)){lit_flags.carelessClear(lit);} asrt(lit_flags.allCleared());

	if(unt){
		if(vals.litTrue(a)){ return false;}
		unit_clses	.push_back(a.flipped());
		vals		.setValue(a.flipped());
	}

	if(eq){
		if(a<b){lit_ifo[b].setReplBy(a); lit_ifo[b.flipped()].setReplBy(a.flipped());}
		else{	lit_ifo[a].setReplBy(b); lit_ifo[a.flipped()].setReplBy(b.flipped());}
	}

	return true;
}*/


/*
bool SimpClses::asrtBinsSorted(const Literal& a) const{
	Iter<Literal> it = getBinsOf(a);
	Literal prev = FALSE_LIT; Literal curr;
	while(it.getNext(curr)){
		if(prev >= curr) return false;
		prev = curr;
	}
	return true;
}*/

/*
ECP	SimpClses::addOrigCls(const ClsP& cp, const Cls& cl,const sig_t& sig,const Literal& w1, const Literal& w2){
		asrt(cls_sig.size()==eclses.size());
		ECP ret(cls_sig.size());
		cls_sig		.push_back(sig);
		eclses		.push_back(ExtCls(cl.litsCnt(),cl.isLearnt(),false,cp));
		cdcl_bkp		.push_back(ClsBkp(ret,cl,w1,w2));
		Literal* lits	= getLitsP(cp);
		lits[0] 		= w1;
		lits[1]		= w2;
		return ret;
}

ECP	SimpClses::addNewCls(uint32_t size, const bool learnt, const bool temp, const Ptr32& lp,const sig_t& sig){
		ECP ret(cls_sig.size());
		cls_sig		.push_back(sig);
		eclses		.push_back(ExtCls(size,learnt,temp,lp));
		cdcl_bkp		.push_back(ClsBkp(ret,size));
		return ret;
}

void	SimpClses::bringLitTo(Literal* lits, const uint32_t& sz, const Literal& cand, const uint32_t& new_pos){
	for(uint32_t i=0;i!=sz;++i){
		if(lits[i]==cand){
			lits[i]		= lits[new_pos];
			lits[new_pos]	= cand;
			return;}
	}
}


void SimpClses::getSimplified(Watching& w, const VarVal& vals){
	w.enlarge(vals.count() << 1);

	glob.stats.nbr_lcls 		= 0;
	glob.stats.nbr_bin_clses 	= 0;
	glob.stats.nbr_lits			= 0;

	for(uint32_t c=0; c!=cdcl_bkp.size();++c){
		ClsBkp& cbk = cdcl_bkp[c];
		if(cbk.isBin()){
			if(	!glob.isElim(cbk.w1.toVar()) 	&& !glob.isElim(cbk.w2.toVar())
			&& !	vals.litKnwn(cbk.w1) 		&& !vals.litKnwn(cbk.w2) ){
				w.clsCreateBinary(cbk.w1,cbk.w2);}
			continue;
		}

		ExtCls& ec = getECls(cbk.eclp);
		if(ec.isWaste()){
			if(!ec.isFreed()){
				glob.silentFree(ec.litsp,cbk.ch.litsCnt());
				ec.freed = 1;
			}
			continue;
		}

		Literal* lits 		= getLitsP(ec.litsp);
		const uint32_t sz	= ec.size(); asrt(cbk.ch.litsCnt() >= sz); asrt(sz>2);

		if(!cbk.isNew()){ //take original watchers
			bringLitTo(lits,sz,cbk.w1,0);
			bringLitTo(lits,sz,cbk.w2,1);
		}

		Literal w1 = lits[0]; Literal w2 = lits[1];
		Cls& cls  = glob.getCls(ec.litsp); asrt((void*)lits == (void*)&cls);
		cls.setHeader(cbk.ch,w1,w2);
		cls.setLearnt(ec.isLearnt());//might have been changed

		if(cbk.ch.litsCnt() > sz){ //clause might have been shrinked
			glob.silentShrink(ec.litsp,cbk.ch.litsCnt()-sz);}

		asrt(cls.litsCnt()==sz);

		w.clsAddWtched(w1,w2,ec.litsp);
		w.clsAddWtched(w2,w1,ec.litsp);
		inc_stat(lits,sz);
		inc_stat(lcls,1);
	}
}
*/


#endif /* CLSINFO_HPP_ */
